/* ==== sig.c =======================================================
 * Copyright (c) 1994 by Chris Provenzano, proven@mit.edu
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by Chris Provenzano.
 * 4. The name of Chris Provenzano may not be used to endorse or promote 
 *	  products derived from this software without specific prior written
 *	  permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CHRIS PROVENZANO ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL CHRIS PROVENZANO BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 *
 * Description : All the thread signal functions.
 *
 *  1.32 94/06/12 proven
 *      -Started coding this file.
 */

#ifndef lint
static const char rcsid[] = "$Id: sig.c,v 1.52 1994/10/19 03:13:44 proven Exp $";
#endif

#include <pthread.h>
#include <signal.h>

extern void sig_handler_real();

struct pthread * pthread_sigwait;
sigset_t __sigpending;

struct pthread_sigvec {
	void 							(*vector)();
	enum pthread_sig_flags {
		PTHREAD_SIG_IGN,
		PTHREAD_SIG_DFL,
	}								flags;
} pthread_sigvec[SIGMAX];

/* ==========================================================================
 * pthread_sig_register()
 *
 * Assumes the kernel is locked.
 */
int pthread_sig_register(int sig)
{
	struct pthread ** pthread_ptr, * pthread;
	int ret; 

	/* Check waiting threads for delivery */
	for (pthread_ptr = &pthread_sigwait; (*pthread_ptr);
	  pthread_ptr = &((*pthread_ptr)->next)) {
		if (sigismember((*pthread_ptr)->sigwait, sig)) {
			pthread_prio_queue_enq(pthread_current_prio_queue, (*pthread_ptr));
			ret = (*pthread_ptr)->pthread_priority;
			*(int *)((*pthread_ptr)->ret) = sig;
			(*pthread_ptr)->state = PS_RUNNING;

			*pthread_ptr = (*pthread_ptr)->next;
			return(ret);
		}
	}

	/* Check current running thread */
	if (!sigismember(&pthread_run->sigmask, sig)) {
		sigaddset(&pthread_run->sigpending, sig);
		return(0);
	}

	/* Check any running thread */
	for (pthread = pthread_current_prio_queue->next;
	  pthread; pthread = pthread->next) {
		if (!sigismember(&pthread->sigmask, sig)) {
			sigaddset(&pthread_run->sigpending, sig);
			return(0);
		}
	}

	/* Check any thread */
	for (pthread = pthread_link_list; pthread; pthread = pthread->pll) {
		if (!sigismember(&pthread->sigmask, sig)) {
			sigaddset(&pthread->sigpending, sig);
			return(0);
		}
	}

	sigaddset(&__sigpending, sig);
}

/* ==========================================================================
 * pthread_sig_process()
 *
 * Assumes the kernel is locked.
 */
int pthread_sig_process()
{
	int i;

	for (i = 1; i < SIGMAX; i++) {
		if (sigismember(&(pthread_run->sigpending), i)) {
			if (! sigismember(&(pthread_run->sigmask), i)) {
				sigdelset(&(pthread_run->sigpending), i);
				pthread_run->sigcount--;

				if (pthread_sigvec[i].flags & PTHREAD_SIG_IGN) {
					continue;
				} 
				if (pthread_sigvec[i].flags & PTHREAD_SIG_DFL) {
					/* Set the signal handler to default */
					signal(i, SIG_DFL);
					raise(i);
					signal(i, sig_handler_real);
					continue;
				} 

				sigaddset(&(pthread_run->sigmask), i);

				
				/* Allow interrupts during a signal */
				pthread_sched_resume();
				pthread_sigvec[i].vector(i);
				pthread_sched_prevent();

				sigdelset(&(pthread_run->sigmask), i);
			}
		}
	}
}

/* ==========================================================================
 * sigprocmask()
 *
 * It is unclear wheather this call should be implemented as an atomic
 * operation. The resulting mask could be wrong if in the signal
 * handler the thread calls sigprocmask for any signal other than the
 * signal the handler is dealing with.
 */
int sigprocmask(int how, const sigset_t *set, sigset_t * oset)
{
	int i;

	if (oset) {
		sigemptyset(oset);
		for (i = 1; i < SIGMAX; i++) {
			if (sigismember(&(pthread_run->sigmask), i)) {
				sigaddset(oset, i);
			}
		}
	}

	switch(how) {
	case SIG_BLOCK:
		for (i = 1; i < SIGMAX; i++) {
			if (sigismember(set, i)) {
				sigaddset(&(pthread_run->sigmask), i);
			}
		}
		break;
	case SIG_UNBLOCK:
		pthread_sched_prevent();
		for (i = 1; i < SIGMAX; i++) {
			if (sigismember(set, i)) {
				sigdelset(&(pthread_run->sigmask), i);
				if (sigismember(&__sigpending, i)) {
					sigaddset(&(pthread_run->sigpending), i);
					sigdelset(&__sigpending, i);
					pthread_run->sigcount++;
				}
			}
		}
		pthread_sched_resume();
		break;
	case SIG_SETMASK:
		sigfillset(&(pthread_run->sigmask));
		pthread_sched_prevent();
		for (i = 1; i < SIGMAX; i++) {
			if (! sigismember(set, i)) {
				sigdelset(&(pthread_run->sigmask), i);
				if (sigismember(&__sigpending, i)) {
					sigaddset(&(pthread_run->sigpending), i);
					sigdelset(&__sigpending, i);
					pthread_run->sigcount++;
				}
			}
		}
		pthread_sched_resume();
		break;
	default:
		SET_ERRNO(EINVAL);
		return(NOTOK);
	}
	return(OK);
}

/* ==========================================================================
 * sigwait()
 */
int sigwait(const sigset_t * set, int * sig)
{
	int i;

	/* Check that sig is valid */
	*sig = 0;

	pthread_sched_prevent();
	for (i = 1; i < SIGMAX; i++) {
		if (sigismember(set, i)) {
			/* Check personal signals */
			if (sigismember(&(pthread_run->sigpending), i)) {
				sigdelset(&(pthread_run->sigpending), i);
				pthread_sched_resume();
				*sig = i;
				return(OK);
			}
			/* Check kernel signals */
			if (sigismember(&__sigpending, i)) {
				sigdelset(&__sigpending, i);
				pthread_sched_resume();
				*sig = i;
				return(OK);
			}
		}
	}

	/* No pending signals, wait for one */
	pthread_run->next = pthread_sigwait;
	pthread_sigwait = pthread_run;
	pthread_run->sigwait = set;
	pthread_run->ret = sig;

	pthread_resched_resume(PS_SIGWAIT);
	return(OK);
}
